from onegov.activity.collections.activity import ActivityCollection
from onegov.activity.collections.booking import BookingCollection
from onegov.activity.collections.occasion import OccasionCollection

__all__ = ['ActivityCollection', 'BookingCollection', 'OccasionCollection']



""" Contains upgrade tasks that are executed when the application is being
upgraded on the server. See :class:`onegov.core.upgrade.upgrade_task`.

"""

from onegov.activity.models import Occasion
from onegov.core.upgrade import upgrade_task
from psycopg2.extras import NumericRange
from sqlalchemy import Column, Text
from sqlalchemy.dialects.postgresql import INT4RANGE


@upgrade_task('Redesign occasion table')
def redesign_occasion_table(context):
    context.operations.drop_column('occasions', 'booking_start')
    context.operations.drop_column('occasions', 'min_age')
    context.operations.drop_column('occasions', 'max_age')
    context.operations.drop_column('occasions', 'spots')

    context.operations.add_column('occasions', Column(
        'note', Text, nullable=True))

    context.operations.add_column('occasions', Column(
        'age', INT4RANGE, nullable=True))

    context.operations.add_column('occasions', Column(
        'spots', INT4RANGE, nullable=True))

    for occasion in context.session.query(Occasion).all():
        occasion.age = NumericRange(6, 17, bounds='[]')
        occasion.spots = NumericRange(0, 10, bounds='[]')

    context.session.flush()

    context.operations.alter_column('occasions', 'spots', nullable=False)
    context.operations.alter_column('occasions', 'age', nullable=False)


@upgrade_task('Make occasion location optional')
def make_occasion_location_optional(context):
    context.operations.alter_column('occasions', 'location', nullable=True)


@upgrade_task('Ensure occasions/bookings cannot be orphaned')
def ensure_occasions_bookings_cannot_be_orphaned(context):
    context.operations.alter_column('occasions', 'activity_id', nullable=False)
    context.operations.alter_column('bookings', 'occasion_id', nullable=False)



from onegov.core.orm import Base
from onegov.core.orm.mixins import (
    content_property,
    ContentMixin,
    meta_property,
    TimestampMixin,
)
from onegov.core.orm.types import UUID
from onegov.core.utils import normalize_for_url
from onegov.user import User
from sqlalchemy import Column, Enum, Text, ForeignKey
from sqlalchemy.dialects.postgresql import HSTORE
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import relationship
from sqlalchemy_utils import observes
from uuid import uuid4


# Note, a database migration is needed if these states are changed
ACTIVITY_STATES = ('preview', 'proposed', 'accepted', 'denied', 'archived')


class Activity(Base, ContentMixin, TimestampMixin):
    """ Describes an activity that is made available to participants on
    certain occasions (i.e. dates).

    The activity describes the what's going on, the occasion describes when
    and with whom.

    """

    __tablename__ = 'activities'

    #: An internal id for references (not public)
    id = Column(UUID, primary_key=True, default=uuid4)

    #: A nice id for the url, readable by humans
    name = Column(Text, nullable=False, unique=True)

    #: The title of the activity
    title = Column(Text, nullable=False)

    #: The normalized title for sorting
    order = Column(Text, nullable=False, index=True)

    #: Describes the activity briefly
    lead = meta_property('lead')

    #: Describes the activity in detail
    text = content_property('text')

    #: Tags/Categories of the activity
    _tags = Column(MutableDict.as_mutable(HSTORE), nullable=True, name='tags')

    #: The user to which this activity belongs to (organiser)
    username = Column(Text, ForeignKey(User.username), nullable=False)

    #: Access the user linked to this activity
    user = relationship('User')

    #: The occasions linked to this activity
    occasions = relationship(
        'Occasion',
        order_by='Occasion.start',
        backref='activity'
    )

    #: the type of the item, this can be used to create custom polymorphic
    #: subclasses of this class. See
    #: `<http://docs.sqlalchemy.org/en/improve_toc/\
    #: orm/extensions/declarative/inheritance.html>`_.
    type = Column(Text, nullable=True)

    #: the state of the activity
    state = Column(
        Enum(*ACTIVITY_STATES, name='activity_state'),
        nullable=False,
        default='preview'
    )

    __mapper_args__ = {
        'polymorphic_on': 'type',
        'order_by': order,
    }

    @observes('title')
    def title_observer(self, title):
        self.order = normalize_for_url(title)

    @property
    def tags(self):
        return set(self._tags.keys()) if self._tags else set()

    @tags.setter
    def tags(self, value):
        self._tags = {k: '' for k in value} if value else None

    def propose(self):
        assert self.state == 'preview'
        self.state = 'proposed'

        return self

    def accept(self):
        assert self.state == 'proposed'
        self.state = 'accepted'

        return self

    def deny(self):
        assert self.state == 'proposed'
        self.state = 'denied'

        return self

    def archive(self):
        assert self.state == 'accepted'
        self.state = 'archived'

        return self



import string

from random import SystemRandom


GROUP_CODE_CHARS = string.ascii_uppercase + string.digits
GROUP_CODE_LENGTH = 10


def random_group_code():
    random = SystemRandom()

    return ''.join(
        random.choice(GROUP_CODE_CHARS)
        for _ in range(GROUP_CODE_LENGTH)
    )



from onegov.activity.models import Occasion
from onegov.core.collection import GenericCollection
from psycopg2.extras import NumericRange
from sedate import standardize_date


class OccasionCollection(GenericCollection):

    @property
    def model_class(self):
        return Occasion

    @staticmethod
    def to_half_open_interval(lower, upper):
        """ Postgres coerces ranges internally to be half-open in an effort
        to canonize these ranges. This function does the same by taking
        a closed interval and turning it into a half-open interval of
        the NumericRange type.

        """
        return NumericRange(lower, upper + 1, bounds='[)')

    def add(self, activity, start, end, timezone,
            location=None, age=None, spots=None, note=None):

        start = standardize_date(start, timezone)
        end = standardize_date(end, timezone)

        return super().add(
            start=start,
            end=end,
            timezone=timezone,
            location=location,
            age=age and self.to_half_open_interval(*age),
            spots=spots and self.to_half_open_interval(*spots),
            note=note,
            activity_id=activity.id
        )



from onegov.activity.models import Booking
from onegov.core.collection import GenericCollection


class BookingCollection(GenericCollection):

    @property
    def model_class(self):
        return Booking

    def by_user(self, user):
        return self.query().filter(Booking.username == user.username)

    def by_username(self, username):
        return self.query().filter(Booking.username == username)

    def add(self, user, priority=None, group_code=None,
            last_name=None, first_name=None):

        return super().add(
            username=user.username,
            priority=priority,
            group_code=group_code,
            last_name=last_name,
            first_name=first_name
        )



from onegov.activity.models.activity import Activity
from onegov.core.orm import Base
from onegov.core.orm.types import UTCDateTime, UUID
from psycopg2.extras import NumericRange
from sedate import to_timezone
from sqlalchemy import Column, Enum, Text, ForeignKey
from sqlalchemy.dialects.postgresql import INT4RANGE
from sqlalchemy.orm import relationship
from uuid import uuid4


class Occasion(Base):
    """ Describes a single occurrence of an Activity. "Occurence" would have
    been a good word for it too, but that's used by onegov.event.

    So occasion it is.

    """

    __tablename__ = 'occasions'

    #: the public id of this occasion
    id = Column(UUID, primary_key=True, default=uuid4)

    #: Timezone of the occasion
    timezone = Column(Text, nullable=False)

    #: Start date and time of the occasion
    start = Column(UTCDateTime, nullable=False)

    #: End date and time of the event (of the first event if recurring)
    end = Column(UTCDateTime, nullable=False)

    #: Describes the location of the activity
    location = Column(Text, nullable=True)

    #: The expected age of participants
    age = Column(
        INT4RANGE, nullable=False, default=NumericRange(6, 17, bounds='[]'))

    #: The expected number of participants
    spots = Column(
        INT4RANGE, nullable=False, default=NumericRange(0, 10, bounds='[]'))

    #: A note about the occurrence
    note = Column(Text, nullable=True)

    #: The activity this occasion belongs to
    activity_id = Column(UUID, ForeignKey(Activity.id), nullable=False)

    #: The bookings linked to this occasion
    bookings = relationship(
        'Booking',
        order_by='Booking.created',
        backref='occasion'
    )

    #: The state of the occasion
    state = Column(
        Enum('active', 'cancelled', name='occasion_state'),
        nullable=False,
        default='active'
    )

    @property
    def localized_start(self):
        """ The localized version of the start date/time. """

        return to_timezone(self.start, self.timezone)

    @property
    def localized_end(self):
        """ The localized version of the end date/time. """

        return to_timezone(self.end, self.timezone)

    def cancel(self):
        assert self.state == 'active'
        self.state == 'cancelled'
